/* Copyright 2015 Jetperch LLC.  All rights reserved. */

#include "embc/fifo.h"
#include "embc/log.h"
#include "embc/dbc.h"
#include "embc/ec.h"
#include <stdlib.h>
#include <string.h> // memcpy

/*
    int head;
    int tail;
    uint8_t * buffer;
    size_t buffer_size;
*/

int fifo8_alloc(struct fifo8_s ** fifo, fifo_size_t depth) {
    struct fifo8_s * f;
    DBC_ARG_NOT_NULL(fifo);
    DBC_ARG_GT_ZERO(depth);

    uint8_t * buffer = (uint8_t *) calloc(1, depth);
    if (!buffer) {
        return EMBC_ERROR_OUT_OF_MEMORY;
    }

    f = (struct fifo8_s *) calloc(1, sizeof(struct fifo8_s));
    if (!f) {
        free(buffer);
        return EMBC_ERROR_OUT_OF_MEMORY;
    }

    f->head = 0;
    f->tail = 0;
    f->buffer = buffer;
    f->buffer_size = depth;
    *fifo = f;

    return EMBC_SUCCESS;
}

void fifo8_free(struct fifo8_s * fifo) {
    if (fifo) {
        if (fifo->buffer) {
            free(fifo->buffer);
            fifo->buffer = 0;
        }
        free(fifo);
    }
}


int fifo8_write(struct fifo8_s * fifo, uint8_t const * data, fifo_size_t size, fifo_size_t * actual_size) {
    fifo_size_t idx;
    fifo_size_t next;
    fifo_size_t sz = 0;
    int rc = EMBC_SUCCESS;

    DBC_ARG_NOT_NULL(fifo);
    DBC_ARG_NOT_NULL(data);
    DBC_ARG_GT_ZERO(size);
    DBC_ARG_NOT_NULL(actual_size);

    for (idx = 0; idx < size; ++idx) {
        next = fifo->head + 1;
        if (next >= fifo->buffer_size) {
            next = 0;
        }
        if (next == fifo->tail) {
            LOGS_WARN("fifo8_write full");
            rc = EMBC_ERROR_FULL;
            break;
        }
        fifo->buffer[fifo->head] = *data++;
        fifo->head = next;
        ++sz;
    }
    DBC_ASSERT_RWE(sz < fifo->buffer_size, "sz too big");
    *actual_size = sz;
    return rc;
}

int fifo8_write_u8(struct fifo8_s * fifo, uint8_t data) {
    fifo_size_t sz;
    return fifo8_write(fifo, &data, 1, &sz);
}

fifo_size_t fifo8_size(struct fifo8_s * fifo) {
    fifo_size_t sz = fifo->head - fifo->tail;
    if (sz < 0) {
        sz += fifo->buffer_size;
    }
    return sz;
}

int fifo8_is_empty(struct fifo8_s * fifo) {
    if (fifo->head == fifo->tail) {
        return 1;
    }
    return 0;
}

int fifo8_is_full(struct fifo8_s * fifo) {
    if (fifo8_size(fifo) + 1 == fifo->buffer_size) {
        return 1;
    }
    return 0;
}


int fifo8_read(struct fifo8_s * fifo, uint8_t * data, fifo_size_t size, fifo_size_t * actual_size) {
    fifo_size_t idx;
    fifo_size_t next;
    fifo_size_t sz = 0;
    int rc = EMBC_SUCCESS;

    DBC_ARG_NOT_NULL(fifo);
    DBC_ARG_NOT_NULL(data);
    DBC_ARG_GT_ZERO(size);
    DBC_ARG_NOT_NULL(actual_size);

    for (idx = 0; idx < size; ++idx) {
        next = fifo->tail + 1;
        if (next >= fifo->buffer_size) {
            next = 0;
        }
        if (fifo->tail == fifo->head) {
            rc = EMBC_ERROR_EMPTY;
            break;
        }
        *data++ = fifo->buffer[fifo->tail];
        fifo->tail = next;
        ++sz;
    }
    *actual_size = sz;
    DBC_ASSERT_RWE(sz < fifo->buffer_size, "sz too big");
    return rc;
}

int fifo8_read_u8(struct fifo8_s * fifo, uint8_t * data) {
    fifo_size_t sz;
    return fifo8_read(fifo, data, 1, &sz);
}
